from pathlib import Path
from typing import Literal, Optional, Tuple, Union

from rdflib import Dataset, Graph, URIRef

from kurra.utils import load_graph

KNOWN_RDF_FORMATS = Literal["turtle", "longturtle", "xml", "n-triples", "json-ld"]
RDF_FILE_SUFFIXES = {
    "turtle": ".ttl",
    "longturtle": ".ttl",
    "xml": ".rdf",
    "n-triples": ".nt",
    "json-ld": ".jsonld",
}


class FailOnChangeError(Exception):
    """
    This exception is raised when running format and the
    check bool is set to true and the file has resulted in a change.
    """


def get_topbraid_metadata(content: str) -> str:
    """Get the TopBraid Composer metadata at the top of an ontology file."""
    lines = content.split("\n")
    comments = []
    for line in lines:
        if line.startswith("#"):
            comments.append(line)
        else:
            break

    if comments:
        return "\n".join(comments) + "\n"
    else:
        return ""


def do_format(
    content: str, output_format: KNOWN_RDF_FORMATS = "longturtle"
) -> Tuple[str, bool]:
    metadata = get_topbraid_metadata(content)

    graph = load_graph(content)
    new_content = graph.serialize(format=output_format)
    new_content = metadata + new_content
    changed = content != new_content
    return new_content, changed


def format_file(
    file: Path,
    check: bool = False,
    output_format: KNOWN_RDF_FORMATS = "longturtle",
    output_filename: Path = None,
) -> bool:
    if not file.is_file():
        raise ValueError(f"{file} is not a file.")

    path = Path(file).resolve()
    if not path.exists():
        raise FileNotFoundError(f"File not found: {path.absolute()}")

    if output_filename is None:
        output_filename = path.with_suffix(RDF_FILE_SUFFIXES[output_format])

    Path(output_filename).touch(exist_ok=True)

    with open(path, "r", encoding="utf-8") as fread:
        content = fread.read()

        content, changed = do_format(content, output_format)
        if check:
            raise FailOnChangeError(
                f"The file {path} contains changes that can be formatted."
            )
        else:
            # Didn't fail and file has changed, so write to file.
            with open(output_filename, "w", encoding="utf-8") as fwrite:
                fwrite.write(content)

    return changed


def format_rdf(
    path: Path,
    check: bool,
    output_format: KNOWN_RDF_FORMATS = "longturtle",
    output_filename: Path = None,
) -> None:
    path = Path(path).resolve()

    if path.is_dir():
        files = list(path.glob("**/*.ttl"))

        changed_files = []

        for file in files:
            try:
                changed = format_file(
                    file,
                    check,
                    output_format=output_format,
                    output_filename=output_filename,
                )
                if changed:
                    changed_files.append(file)
            except FailOnChangeError as err:
                print(err)
                changed_files.append(file)

        if check and changed_files:
            if changed_files:
                raise FailOnChangeError(
                    f"{len(changed_files)} out of {len(files)} files will change."
                )
            else:
                print(
                    f"{len(changed_files)} out of {len(files)} files will change.",
                )
        else:
            print(
                f"{len(changed_files)} out of {len(files)} files changed.",
            )
    else:
        try:
            format_file(
                path,
                check,
                output_format=output_format,
                output_filename=output_filename,
            )
        except FailOnChangeError as err:
            print(err)


def make_dataset(
    path_str_or_graph: Union[Path, str, Graph], graph_iri: Union[str, URIRef]
) -> Dataset:
    """Returns a given Graph, or string or file of triples, as a Dataset, with the supplied graph IRI"""

    # TODO: make a Dataset from a Graph or Datatset
    # - override option to replace existing graph
    # - set default union graph
    # - set default graph
    if not isinstance(graph_iri, URIRef):
        graph_iri = URIRef(graph_iri)

    g = load_graph(path_str_or_graph)

    d = Dataset()
    for s, p, o in g:
        d.add((s, p, o, graph_iri))

    return d


def export_quads(
    path_str_or_dataset: Union[Path, str, Dataset], destination: Optional[Path] = None
) -> bool | str:
    """Exports a given Dataset, or quads in trig format or a quads file specified by a path, either as
    quads to a string, if no destination is given, or a file, if one is"""
    if isinstance(path_str_or_dataset, Path):
        d = Dataset()
        d.parse(str(path_str_or_dataset))
    elif isinstance(path_str_or_dataset, str):
        d = Dataset()
        d.parse(data=path_str_or_dataset, format="trig")
    else:  # Dataset
        d = path_str_or_dataset

    if destination is not None:
        if Path(destination).is_file():
            d2 = Dataset()
            d2.parse(destination)
            d3 = d + d2
            d3.serialize(format="trig", destination=destination)
        else:
            d.serialize(format="trig", destination=destination)

        return True
    else:
        return d.serialize(format="trig")



from io import StringIO
from pathlib import Path

from rdflib import RDF, Graph, URIRef

from kurra.db import create_dataset, list_dataset


def test_db_create_dataset_by_dataset_name(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}"
    dataset_name = "myds"
    return_value = create_dataset(base_url, dataset_name, http_client=http_client)
    assert return_value == f"Dataset {dataset_name} created at {base_url}."

    result = list_dataset(base_url, http_client=http_client)
    assert f"/{dataset_name}" in list(map(lambda x: x["ds.name"], result))


def test_db_create_dataset_by_config_file(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}"
    dataset_name = "myds"
    config_file = StringIO(
        """PREFIX :          <http://base/#>
PREFIX bibo:      <http://purl.org/ontology/bibo/>
PREFIX dc:        <http://purl.org/dc/elements/1.1/>
PREFIX dcterms:   <http://purl.org/dc/terms/>
PREFIX ex:        <http://example.org/>
PREFIX fuseki:    <http://jena.apache.org/fuseki#>
PREFIX geosparql: <http://jena.apache.org/geosparql#>
PREFIX rdf:       <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs:      <http://www.w3.org/2000/01/rdf-schema#>
PREFIX schema:    <https://schema.org/>
PREFIX skos:      <http://www.w3.org/2004/02/skos/core#>
PREFIX tdb2:      <http://jena.apache.org/2016/tdb#>
PREFIX text:      <http://jena.apache.org/text#>

:service_tdb_all  rdf:type  fuseki:Service;
        rdfs:label       "myds";
        fuseki:dataset   :geosparql_dataset;
        fuseki:endpoint  [ fuseki:name       "query";
                           fuseki:operation  fuseki:query
                         ];
        fuseki:endpoint  [ fuseki:name       "data";
                           fuseki:operation  fuseki:gsp-rw
                         ];
        fuseki:endpoint  [ fuseki:name       "get";
                           fuseki:operation  fuseki:gsp-r
                         ];
        fuseki:endpoint  [ fuseki:operation  fuseki:update ];
        fuseki:endpoint  [ fuseki:operation  fuseki:gsp-rw ];
        fuseki:endpoint  [ fuseki:name       "sparql";
                           fuseki:operation  fuseki:query
                         ];
        fuseki:endpoint  [ fuseki:operation  fuseki:query ];
        fuseki:endpoint  [ fuseki:name       "update";
                           fuseki:operation  fuseki:update
                         ];
        fuseki:name      "myds" .

:geosparql_dataset  rdf:type            geosparql:geosparqlDataset;
        geosparql:applyDefaultGeometry  false;
        geosparql:dataset               :text_dataset;
        geosparql:indexEnabled          true;
        geosparql:indexExpires          "5000,5000,5000";
        geosparql:indexSizes            "-1,-1,-1";
        geosparql:inference             false;
        geosparql:queryRewrite          true;
        geosparql:spatialIndexFile      "/fuseki/databases/myds/spatial.index" .

:text_dataset  rdf:type  text:TextDataset;
        text:dataset  :tdb_dataset_readwrite;
        text:index    :index_lucene .

:tdb_dataset_readwrite
        rdf:type                tdb2:DatasetTDB2;
        tdb2:location           "/fuseki/databases/myds";
        tdb2:unionDefaultGraph  true .

:index_lucene  rdf:type   text:TextIndexLucene;
        text:analyzer     [ rdf:type  text:StandardAnalyzer ];
        text:directory    "/fuseki/databases/myds";
        text:entityMap    :entity_map;
        text:propLists    ( [ text:propListProp  ex:searchFields;
                              text:props         ( schema:headline schema:name bibo:shortTitle dcterms:title dc:title bibo:abstract schema:description dc:description dcterms:description rdfs:label skos:prefLabel skos:altLabel )
                            ]
                          );
        text:storeValues  true .

:entity_map  rdf:type      text:EntityMap;
        text:defaultField  "prefLabel";
        text:entityField   "uri";
        text:graphField    "graph";
        text:langField     "lang";
        text:map           ( [ text:field      "prefLabel";
                               text:predicate  skos:prefLabel
                             ]
                             [ text:field      "altLabel";
                               text:predicate  skos:altLabel
                             ]
                             [ text:field      "notation";
                               text:predicate  skos:notation
                             ]
                             [ text:field      "definition";
                               text:predicate  skos:definition
                             ]
                             [ text:field      "hidden";
                               text:predicate  skos:hiddenLabel
                             ]
                             [ text:field      "rdfslabel";
                               text:predicate  rdfs:label
                             ]
                             [ text:field      "headline";
                               text:predicate  schema:headline
                             ]
                             [ text:field      "name";
                               text:predicate  schema:name
                             ]
                             [ text:field      "shortTitle";
                               text:predicate  bibo:shortTitle
                             ]
                             [ text:field      "dcttitle";
                               text:predicate  dcterms:title
                             ]
                             [ text:field      "dctitle";
                               text:predicate  dc:title
                             ]
                             [ text:field      "abstract";
                               text:predicate  bibo:abstract
                             ]
                             [ text:field      "sdodescription";
                               text:predicate  schema:description
                             ]
                             [ text:field      "dctdescription";
                               text:predicate  dcterms:description
                             ]
                             [ text:field      "dcdescription";
                               text:predicate  dc:description
                             ]
                           );
        text:uidField      "uid" ."""
    )
    return_value = create_dataset(base_url, config_file, http_client=http_client)
    assert (
        return_value
        == f"Dataset {dataset_name} created using assembler config at {base_url}."
    )

    result = list_dataset(base_url, http_client=http_client)
    assert f"/{dataset_name}" in list(map(lambda x: x["ds.name"], result))


def test_db_create_dataset_by_config_file_with_existing_dataset(
    fuseki_container, http_client
):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}"
    current_dir = Path(__file__).parent
    file = current_dir / "config.ttl"
    graph = Graph().parse(file, format="turtle")
    fuseki_service = graph.value(
        None, RDF.type, URIRef("http://jena.apache.org/fuseki#Service")
    )
    dataset_name = graph.value(
        fuseki_service, URIRef("http://jena.apache.org/fuseki#name")
    )
    return_value = create_dataset(base_url, file, http_client=http_client)

    assert (
        return_value
        == f"Dataset {dataset_name} created using assembler config at {base_url}."
    )

    result = list_dataset(base_url, http_client=http_client)
    assert f"/{dataset_name}" in list(map(lambda x: x["ds.name"], result))



from pathlib import Path
from textwrap import dedent

import httpx
from typer.testing import CliRunner

from kurra.cli import app
from kurra.db import upload

runner = CliRunner()


LANG_TEST_VOC = (
    Path(__file__).parent.parent.resolve() / "test_sparql" / "language-test.ttl"
)


def test_query_db(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"

    with httpx.Client() as client:
        upload(SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, client)

    q = dedent("""
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            GRAPH ?g {
                ?c a skos:Concept .
            } 
        }""").replace("\n", "")

    result = runner.invoke(
        app,
        [
            "sparql",
            SPARQL_ENDPOINT,
            q,
        ],
    )
    print("BEFORE")
    print(result)
    print("AFTER")
    # assert result.exit_code == 0


def test_query_file():
    # TODO: work out why this test fails but the direct call works
    # direct call:
    # kurra sparql /Users/nick/work/kurrawong/kurra/tests/test_sparql/language-test.ttl "PREFIX skos: <http://www.w3.org/2004/02/skos/core#> SELECT * WHERE {     ?c a skos:Concept .}"
    q = dedent("""
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            ?c a skos:Concept .
        }""").replace("\n", "")

    result = runner.invoke(
        app,
        ["sparql", LANG_TEST_VOC, q],
    )



import sys
from pathlib import Path
from typing import Annotated

import typer

from kurra.cli.commands.db import upload_command
from kurra.cli.commands.sparql import sparql_command
from kurra.cli.console import console
from kurra.file import (
    RDF_FILE_SUFFIXES,
    FailOnChangeError,
    export_quads,
    format_rdf,
    make_dataset,
)

app = typer.Typer(help="RDF file commands")


@app.command(
    name="format", help="Format RDF files using one of several common RDF formats"
)
def format_command(
    file_or_dir: str = typer.Argument(
        ..., help="The file or directory of RDF files to be formatted"
    ),
    check: bool = typer.Option(
        False,
        "--check",
        "-c",
        help="Check whether files will be formatted without applying the effect.",
    ),
    output_format: str = typer.Option(
        "longturtle",
        "--output-format",
        "-f",
        help=f"Indicate the output RDF format. Available are {list(RDF_FILE_SUFFIXES.keys())}.",
    ),
    output_filename: str = typer.Option(
        None,
        "--output-filename",
        "-o",
        help="the name of the file you want to write the reformatted content to",
    ),
) -> None:
    try:
        format_rdf(file_or_dir, check, output_format, output_filename)
    except FailOnChangeError as err:
        print(err)
        sys.exit(1)


@app.command(name="upload", help="Upload files to a database repository")
def upload_command2(
    path: Path = typer.Argument(
        ..., help="The path of a file or directory to be uploaded."
    ),
    fuseki_url: str = typer.Argument(
        ..., help="Repository SPARQL Endpoint URL. E.g. http://localhost:3030/ds"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    upload_command(path, fuseki_url, username, password, timeout)


@app.command(
    name="quads",
    help="Exports (prints or saves) triples as quads with a given identifier",
)
def quads_command(path_or_str: Path, identifier: str, destination: Path = None):
    r = export_quads(make_dataset(path_or_str, identifier), destination)
    if not destination:
        console.print(r)


@app.command(name="sparql", help="SPARQL queries to local RDF files or a database")
def query_command2(
    path_or_url: Path,
    q: str,
    response_format: str = typer.Option(
        "table",
        "--response-format",
        "-f",
        help="The response format of the SPARQL query. Either 'table' (default) or 'json'",
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    sparql_command(path_or_url, q, response_format, username, password, timeout)



from pathlib import Path

from typer.testing import CliRunner

from kurra.cli import app

runner = CliRunner()


def test_cli_valid():
    SHACL_TEST_DIR = Path(__file__).parent.parent.resolve() / "test_shacl"

    result = runner.invoke(
        app,
        [
            "shacl",
            "validate",
            f"{SHACL_TEST_DIR / 'vocab-valid.ttl'}",
            f"{SHACL_TEST_DIR / 'validator-vocpub-410.ttl'}",
        ],
    )
    assert result.stdout.strip() == "The data is valid"


def test_cli_invalid():
    SHACL_TEST_DIR = Path(__file__).parent.parent.resolve() / "test_shacl"

    result = runner.invoke(
        app,
        [
            "shacl",
            "validate",
            f"{SHACL_TEST_DIR / 'vocab-invalid.ttl'}",
            f"{SHACL_TEST_DIR / 'validator-vocpub-410.ttl'}",
        ],
    )
    assert "The errors are:" in result.stdout



from json import loads

from rdflib import Graph
from rdflib.namespace import RDF, SH
from rdflib.plugins.sparql.processor import SPARQLResult
from rich.table import Table


def format_sparql_response_as_rich_table(response):
    t = Table()

    # ASK
    if not response.get("results"):
        t.add_column("Ask")
        t.add_row(str(response["boolean"]))
    else:  # SELECT
        for x in response["head"]["vars"]:
            t.add_column(x)
        for row in response["results"]["bindings"]:
            cols = []
            for k, v in row.items():
                cols.append(v["value"])
            t.add_row(*tuple(cols))

    return t


def format_sparql_response_as_json(response):
    if isinstance(response, SPARQLResult):
        response = loads(response.serialize(format="json").decode())

    return response


def format_shacl_graph_as_rich_table(g: Graph):
    t = Table(padding=(1, 0))
    t.add_column("No.")
    t.add_column("Error")
    t.add_column("Message")
    errs = 0
    for vr in g.subjects(RDF.type, SH.ValidationResult):
        errs += 1
        t.add_row(
            str(errs),
            g.value(vr, SH.focusNode),
            g.value(vr, SH.resultMessage),
        )

    return t



import json
from pathlib import Path

from rdflib import Graph
from rdflib.compare import isomorphic

from kurra.utils import (
    RenderFormat,
    guess_format_from_data,
    load_graph,
    render_sparql_result,
)


def test_guess_format_from_data():
    s = """
        PREFIX ex: <http://example.com/>
        
        ex:a ex:b ex:c .
        """

    assert guess_format_from_data(s) == "text/turtle"

    s2 = """
        @prefix ex: <http://example.com/> .

        ex:a ex:b ex:c .
        """

    assert guess_format_from_data(s2) == "text/turtle"

    s3 = """
        [
          {
            "@id": "http://example.com/a",
            "http://example.com/b": [
              {
                "@id": "http://example.com/c"
              }
            ]
          }
        ]
        """

    assert guess_format_from_data(s3) == "application/ld+json"

    s4 = """
        <http://example.com/a> <http://example.com/b> <http://example.com/c> .
        """

    assert guess_format_from_data(s4) == "application/n-triples"

    s5 = """
        <?xml version="1.0" encoding="utf-8"?>
        <rdf:RDF
           xmlns:ex="http://example.com/"
           xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        >
          <rdf:Description rdf:about="http://example.com/a">
            <ex:b rdf:resource="http://example.com/c"/>
          </rdf:Description>
        </rdf:RDF>
        """

    assert guess_format_from_data(s5) == "application/rdf+xml"

    # TODO: properly handle detection of HexTuples
    sx = """
        ["http://example.com/a", "http://example.com/b", "http://example.com/c", "globalId", "", ""]
        """

    assert guess_format_from_data(sx) == "application/ld+json"


def test_load_graph():
    g = Graph()
    g.parse(
        data="""
            PREFIX ex: <http://example.com/>
            
            ex:a ex:b ex:c .
            """
    )

    # load from a given Graph
    g2 = load_graph(g)

    assert isomorphic(g2, g)

    # load an RDF file
    g3 = load_graph(Path(__file__).parent / "test_file" / "minimal1.ttl")

    assert isomorphic(g3, g)

    # load data
    g4 = load_graph(
        """
            PREFIX ex: <http://example.com/>
            
            ex:a ex:b ex:c .
            """
    )

    assert isomorphic(g4, g)

    g5 = load_graph(
        "https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl"
    )

    assert len(g5) > 10


def test_load_graph_dir():
    DIR_OF_RDF = Path(__file__).parent / "dir_of_rdf"
    g = Graph()
    g.parse(DIR_OF_RDF / "rdf_1.ttl")
    g.parse(DIR_OF_RDF / "rdf_2.ttl")
    g.parse(DIR_OF_RDF / "rdf_3.ttl")

    g2 = load_graph(DIR_OF_RDF)

    assert len(g2) == len(g)

    g.parse(DIR_OF_RDF / "subdir" / "rdf_4.ttl")

    g3 = load_graph(DIR_OF_RDF, recursive=True)

    assert len(g3) == len(g)


def test_render_sparql_result():
    # simple Python
    r1 = {
        "head": {"vars": ["iri", "value"]},
        "results": {
            "bindings": [
                {
                    "iri": {
                        "type": "uri",
                        "value": "https://linked.data.gov.au/dataset/qld-addr/address/605bf8e7-315a-562b-af4c-16a870732daf",
                    },
                    "value": {
                        "type": "literal",
                        "value": "72 Yundah Street, Shorncliffe, Queensland, Australia",
                    },
                },
                {
                    "iri": {
                        "type": "uri",
                        "value": "https://linked.data.gov.au/dataset/qld-addr/address/005fd678-6957-5953-975b-983515d3c145",
                    },
                    "value": {
                        "type": "literal",
                        "value": "104 Yundah Street, Shorncliffe, Queensland, Australia",
                    },
                },
            ]
        },
    }

    assert "| --- | --- |" in render_sparql_result(r1)

    # simple JSON
    r2 = """
{
  "head": {
    "vars": [
      "s",
      "p",
      "o"
    ]
  },
  "results": {
    "bindings": [
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
        },
        "o": {
          "type": "uri",
          "value": "http://www.w3.org/2004/02/skos/core#Concept"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://purl.org/linked-data/registry#status"
        },
        "o": {
          "type": "uri",
          "value": "http://def.isotc211.org/19135/-1/2015/CoreModel/code/RE_ItemStatus/stable"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/2000/01/rdf-schema#isDefinedBy"
        },
        "o": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/2004/02/skos/core#definition"
        },
        "o": {
          "type": "literal",
          "xml:lang": "en",
          "value": "Missing"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/2004/02/skos/core#historyNote"
        },
        "o": {
          "type": "literal",
          "xml:lang": "en",
          "value": "Presented in the original standard's codelist"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/2004/02/skos/core#inScheme"
        },
        "o": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/2004/02/skos/core#prefLabel"
        },
        "o": {
          "type": "literal",
          "xml:lang": "en",
          "value": "accepted"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "http://www.w3.org/2004/02/skos/core#topConceptOf"
        },
        "o": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted"
        },
        "p": {
          "type": "uri",
          "value": "https://schema.org/identifier"
        },
        "o": {
          "type": "literal",
          "datatype": "http://www.w3.org/2001/XMLSchema#token",
          "value": "accepted"
        }
      },
      {
        "s": {
          "type": "uri",
          "value": "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
        },
        "p": {
          "type": "uri",
          "value": "https://schema.org/description"
        },
        "o": {
          "type": "literal",
          "value": "The subject is an instance of a class."
        }
      }
    ]
  }
}        
        """
    assert "| --- | --- | --- |" in render_sparql_result(r2)

    # OPTIONAL values / no values
    # multiple language literals
    r3 = """
{
  "head": {
    "vars": [
      "cs",
      "c",
      "pl",
      "al"
    ]
  },
  "results": {
    "bindings": [
      {
        "cs": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test"
        },
        "c": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test/en-variant"
        },
        "pl": {
          "type": "literal",
          "xml:lang": "eng",
          "value": "English prefLabel eng"
        }
      },
      {
        "cs": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test"
        },
        "c": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test/en-variant"
        },
        "pl": {
          "type": "literal",
          "xml:lang": "en-AU",
          "value": "English prefLabel en-au"
        }
      },
      {
        "cs": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test"
        },
        "c": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test/altlabels"
        },
        "pl": {
          "type": "literal",
          "xml:lang": "en",
          "value": "English prefLabel"
        },
        "al": {
          "type": "literal",
          "xml:lang": "pl",
          "value": "Polski prefLabel"
        }
      },
      {
        "cs": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test"
        },
        "c": {
          "type": "uri",
          "value": "https://example.com/demo-vocabs/language-test/altlabels"
        },
        "pl": {
          "type": "literal",
          "xml:lang": "en",
          "value": "English prefLabel"
        },
        "al": {
          "type": "literal",
          "xml:lang": "ar",
          "value": "العربيةالعلامة المفضلة"
        }
      }
    ]
  }
}        
        """

    assert "| --- | --- | --- | --- |" in render_sparql_result(r3)

    r4 = """
        {
          "head": {},
          "boolean": true
        }
        """

    assert render_sparql_result(r4) == "True"

    r5 = """
        {
          "head": {},
          "boolean": false
        }        
        """

    assert render_sparql_result(r5) == "False"

    r6 = Graph().parse(
        data="""
            <https://pid.geoscience.gov.au/def/voc/ga/BoreholeStatus/completed> <http://schema.org/name> "completed"@en .
            <http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/accepted> <http://schema.org/name> "accepted"@en .
            <http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/completed> <http://schema.org/name> "completed"@en .            
            """,
        format="turtle",
    )

    assert render_sparql_result(r6).startswith("```")

    r7 = """
        {
          "head": {},
          "boolean": false
        }        
        """

    expected = {"head": {}, "boolean": False}

    assert json.loads(render_sparql_result(r7, RenderFormat.json)) == expected

    assert (
        '"@id": "http://def.isotc211.org/19115/-1/2014/IdentificationInformation/code/MD_ProgressCode/completed"'
        in str(render_sparql_result(r6, RenderFormat.json))
    )



from pathlib import Path
from typing import Annotated
import httpx
import typer

from kurra.cli.console import console
from kurra.cli.utils import (
    format_sparql_response_as_json,
    format_sparql_response_as_rich_table,
)
from kurra.sparql import query

app = typer.Typer()


@app.command(name="sparql", help="SPARQL queries to local RDF files or a database")
def sparql_command(
    path_or_url: Path,
    q: str,
    response_format: str = typer.Option(
        "table",
        "--response-format",
        "-f",
        help="The response format of the SPARQL query. Either 'table' (default) or 'json'",
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    """SPARQL queries a local file or SPARQL Endpoint"""
    if str(path_or_url).startswith("http"):
        path_or_url = str(path_or_url).replace(":/", "://")

    auth = (
        (username, password) if username is not None and password is not None else None
    )
    with httpx.Client(auth=auth, timeout=timeout) as http_client:
        r = query(path_or_url, q, http_client, return_python=True)

        if response_format == "table":
            console.print(format_sparql_response_as_rich_table(r))
        else:
            console.print(format_sparql_response_as_json(r))




import pytest

from kurra.db import FusekiError, delete_dataset


def test_db_delete_dataset(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}"
    dataset_name = "ds"
    return_value = delete_dataset(base_url, dataset_name, http_client)
    assert return_value == f"Dataset {dataset_name} deleted."


def test_db_delete_dataset_non_existent(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}"
    dataset_name = "non-existent"

    with pytest.raises(FusekiError) as exc_info:
        delete_dataset(base_url, dataset_name, http_client)

    assert f"Failed to delete dataset '{dataset_name}'" in exc_info.value.message



import httpx
import json
from enum import Enum
from pathlib import Path
from typing import Union
import pickle
from rdflib import Graph


class RenderFormat(str, Enum):
    original = "original"
    json = "json"
    markdown = "markdown"


class SPARQL_RESULTS_MEDIA_TYPES(str, Enum):
    json = "application/json"
    turtle = "text/turtle"
    jsonld = "application/ld+json"


def guess_format_from_data(rdf: str) -> str | None:
    if rdf is not None:
        rdf = rdf.strip()
        if rdf.startswith("PREFIX") or rdf.startswith("@prefix"):
            return "text/turtle"
        elif rdf.startswith("{") or rdf.startswith("["):
            return "application/ld+json"
        elif rdf.startswith("<?xml") or rdf.startswith("<rdf"):
            return "application/rdf+xml"
        elif rdf.startswith("<http"):
            return "application/n-triples"
        else:
            return "application/n-triples"
    else:
        return None


def load_graph(graph_path_or_str: Union[Graph, Path, str], recursive=False) -> Graph:
    """
    Presents an RDFLib Graph object from a pre-existing Graph, a pickle file, an RDF file or directory of files or RDF
    data in a string
    """
    # Pre-existing Graph
    if isinstance(graph_path_or_str, Graph):
        return graph_path_or_str

    # Pickle file
    if isinstance(graph_path_or_str, Path):
        if graph_path_or_str.is_file():
            pkl_path = graph_path_or_str.with_suffix(".pkl")
            if pkl_path.is_file():
                return pickle.load(open(pkl_path, "rb"))

    # Serialized RDF file or dir of files
    if isinstance(graph_path_or_str, Path):
        if Path(graph_path_or_str).is_file():
            return Graph().parse(str(graph_path_or_str))
        elif Path(graph_path_or_str).is_dir():
            g = Graph()
            if recursive:
                gl = Path(graph_path_or_str).rglob("*.ttl")
            else:
                gl = Path(graph_path_or_str).glob("*.ttl")
            for f in gl:
                if f.is_file():
                    g.parse(f)
            return g

    # A remote file via HTTP
    elif isinstance(graph_path_or_str, str) and graph_path_or_str.startswith("http"):
        return Graph().parse(graph_path_or_str)

    # RDF data in a string
    else:
        return Graph().parse(
            data=graph_path_or_str,
            format=guess_format_from_data(graph_path_or_str),
        )


def render_sparql_result(
    r: dict | str | Graph, rf: RenderFormat = RenderFormat.markdown
) -> str:
    """Renders a SPARQL result in a given render format"""
    if rf == RenderFormat.original:
        return r

    elif rf == RenderFormat.json:
        if isinstance(r, dict):
            return json.dumps(r, indent=4)
        elif isinstance(r, str):
            return json.dumps(json.loads(r), indent=4)
        elif isinstance(r, Graph):
            return r.serialize(format="json-ld", indent=4)

    elif rf == RenderFormat.markdown:
        if isinstance(r, Graph):  # CONSTRUCT: RDF GRaph
            output = "```turtle\n" + r.serialize(format="longturtle") + "```\n"
        else:  # SELECT or ASK: Python dict or JSON

            def render_sparql_value(v: dict) -> str:
                # TODO: handle v["datatype"]
                if v is None:
                    return ""
                elif v["type"] == "uri":
                    return f"[{v['value'].split('/')[-1].split('#')[-1]}]({v['value']})"
                elif v["type"] == "literal":
                    return v["value"]
                elif v["type"] == "bnode":
                    return f"BN: {v['value']:>6}"

            if isinstance(r, str):
                r = json.loads(r)

            output = ""
            header = ["", ""]
            body = []

            if r.get("head") is not None:
                # SELECT
                if r["head"].get("vars") is not None:
                    for col in r["head"]["vars"]:
                        header[0] += f"{col} | "
                        header[1] += f"--- | "
                    output = (
                        "| " + header[0].strip() + "\n| " + header[1].strip() + "\n"
                    )

            if r.get("results"):
                if r["results"].get("bindings"):
                    for row in r["results"]["bindings"]:
                        row_cols = []
                        for k in r["head"]["vars"]:
                            v = row.get(k)
                            if v is not None:
                                # ignore the k
                                row_cols.append(render_sparql_value(v))
                            else:
                                row_cols.append("")
                        body.append(" | ".join(row_cols))

                output += "\n| ".join(body) + " |\n"

            if r.get("boolean") is not None:
                output = str(bool(r.get("boolean")))

        return output


def make_httpx_client(
    sparql_username: str = None,
    sparql_password: str = None,
):
    auth = None
    if sparql_username:
        if sparql_password:
            auth = httpx.BasicAuth(sparql_username, sparql_password)
    return httpx.Client(auth=auth)



from kurra.cli.app import app
from kurra.cli.commands.db import app as db_app
from kurra.cli.commands.file import app as file_app
from kurra.cli.commands.shacl import app as shacl_app
from kurra.cli.commands.sparql import app as sparql_app

app.add_typer(db_app, name="db")
app.add_typer(file_app, name="file")
app.add_typer(shacl_app, name="shacl")
app.add_typer(sparql_app)



from pathlib import Path

from kurra.shacl import validate
from kurra.utils import load_graph

SHACL_TEST_DIR = Path(__file__).parent.resolve()


def test_validate_simple():
    shacl_graph = load_graph(SHACL_TEST_DIR / "validator-vocpub-410.ttl")

    data_file = SHACL_TEST_DIR / "vocab-valid.ttl"
    valid, g, txt = validate(data_file, shacl_graph)
    assert valid

    data_file2 = SHACL_TEST_DIR / "vocab-invalid.ttl"
    valid2, g2, txt2 = validate(data_file2, shacl_graph)
    assert not valid2

    data_file3 = SHACL_TEST_DIR / "vocab-invalid2.ttl"
    valid3, g3, txt3 = validate(data_file3, shacl_graph)
    assert not valid3



from pathlib import Path

import httpx
import pytest
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs

FUSEKI_IMAGE = "ghcr.io/kurrawong/fuseki-geosparql:git-main-e642d849"


@pytest.fixture(scope="function")
def fuseki_container(request: pytest.FixtureRequest):
    container = DockerContainer(FUSEKI_IMAGE)
    container.with_volume_mapping(
        str(Path(__file__).parent / "shiro.ini"),
        "/fuseki/shiro.ini",
    )
    container.with_volume_mapping(
        str(Path(__file__).parent / "config.ttl"),
        "/fuseki/config.ttl",
    )
    container.with_exposed_ports(3030)
    container.start()
    wait_for_logs(container, "Started")

    def cleanup():
        container.stop()

    request.addfinalizer(cleanup)
    return container


@pytest.fixture(scope="function")
def http_client(request: pytest.FixtureRequest):
    _http_client = httpx.Client(auth=("admin", "admin"))

    def cleanup():
        _http_client.close()

    request.addfinalizer(cleanup)
    return _http_client



from typing import Annotated

import typer

from kurra import __version__
from kurra.cli.console import console

app = typer.Typer(
    invoke_without_command=True,
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
    add_completion=False,
)


@app.callback(invoke_without_command=True)
def main(
    version: Annotated[bool, typer.Option("--version", "-v", is_eager=True)] = False,
):
    """Main callback for the CLI app"""
    if version:
        console.print(__version__)
        raise typer.Exit()



from typer.testing import CliRunner

from kurra.cli import app

runner = CliRunner()


def test_cli_db_upload_file(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    result = runner.invoke(
        app,
        [
            "db",
            "upload",
            "tests/test_fuseki/minimal1.ttl",
            f"http://localhost:{port}/ds",
        ],
    )
    assert result.exit_code == 0



import subprocess
import warnings
from pathlib import Path

from rdflib import Dataset, Graph, URIRef

from kurra.file import export_quads, format_file, make_dataset
from kurra.utils import load_graph
from textwrap import dedent


def test_format_rdf_one():
    input_file = Path(__file__).parent / "minimal1.ttl"
    output_file = Path(__file__).parent / "minimal1-out.ttl"
    comparison = """PREFIX ex: <http://example.com/>

ex:a
    ex:b ex:c ;
."""

    format_file(input_file, False, output_filename=output_file)
    output_file_text = output_file.read_text().strip().replace("ns1", "ex")

    assert output_file_text == comparison

    output_file.unlink(missing_ok=True)

    input_file = Path(__file__).parent / "minimal1b.ttl"
    output_file = Path(__file__).parent / "minimal1b-out.ttl"
    # same comparison data

    format_file(input_file, False, output_filename=output_file)
    output_file_text = output_file.read_text().strip().replace("ns1", "ex")

    assert output_file_text == comparison

    output_file.unlink(missing_ok=True)


def test_format_cli():
    subprocess.check_output(
        [
            "kurra",
            "file",
            "format",
            "--output-format",
            "json-ld",
            "tests/test_file/minimal1.ttl",
        ]
    )

    comparison = """[
  {
    "@id": "http://example.com/a",
    "http://example.com/b": [
      {
        "@id": "http://example.com/c"
      }
    ]
  }
]"""

    output_file = Path(__file__).parent / "minimal1.jsonld"

    assert open(output_file).read() == comparison

    Path.unlink(output_file)


def test_make_dataset():
    g = load_graph(
        """
        PREFIX ex: <http://example.com/>
        
        ex:a ex:b ex:c . 
        """
    )

    d = make_dataset(g, "http://graph.com/a")

    for t in d.quads():
        assert t[0] == URIRef("http://example.com/a")
        assert t[1] == URIRef("http://example.com/b")
        assert t[2] == URIRef("http://example.com/c")
        assert t[3] == URIRef("http://graph.com/a")


def test_export_quads():
    g = Graph()
    g.parse(
        data="""
            PREFIX ex: <http://example.com/>

            ex:a ex:b ex:c . 
            """,
        format="turtle",
    )

    d = make_dataset(g, "http://graph.com/a")

    qds = export_quads(d)

    warnings.filterwarnings(
        "ignore", category=DeprecationWarning
    )  # ignore RDFLib's ConjunctiveGraph warning
    d2 = Dataset()
    d2.parse(data=qds, format="trig")

    for t in d.quads():
        assert t[0] == URIRef("http://example.com/a")
        assert t[1] == URIRef("http://example.com/b")
        assert t[2] == URIRef("http://example.com/c")
        assert t[3] == URIRef("http://graph.com/a")


def test_sparql():
    # x = subprocess.check_output(
    #     ["kurra", "query", "--f", "table", "tests/minimal1.ttl", "ASK { ?s ?p ?o}"]
    # )
    #
    # print(x)
    pass


# TODO: NC: solution not found yet to carry the namespaces from the graph through to serialization
def test_prefixes():
    input_file = Path(__file__).parent / "prefixes-test.ttl"
    output_file = Path(__file__).parent / "prefixes-test-out.ttl"
    comparison = dedent(
        """
        PREFIX ex: <http://example.com/>
        PREFIX other: <http://other.com/>
        PREFIX sss: <https://schema.org/>
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        
        ex:a
            ex:b other:c ;
        .
        
        ex:x 
            a skos:ConceptScheme ; 
            ex:b sss:y ;
        .
        """
    )

    format_file(input_file, False, output_filename=output_file)

    # assert output_file == comparison

    output_file.unlink(missing_ok=True)


from pathlib import Path
from typing import Annotated

import httpx
import typer
from rich.progress import track

from kurra.cli.commands.sparql import sparql_command
from kurra.cli.console import console
from kurra.db import (
    FusekiError,
    clear_graph,
    create_dataset,
    delete_dataset,
    list_dataset,
    suffix_map,
    upload,
)

app = typer.Typer(help="RDF database commands. Currently only Fuseki is supported")

dataset_type_options = ["mem", "tdb", "tdb1", "tdb2"]


@app.command(name="list", help="Get the list of database repositories")
def repository_list_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as client:
        try:
            result = list_dataset(fuseki_url, client)
            console.print(result)
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to list repositories at {fuseki_url}."
            )
            raise err


@app.command(
    name="create",
    help="Create a new database repository. Provide either the dataset name and optionally the dataset type or provide the assembler file with --config.",
)
def repository_create_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    dataset_name: str | None = typer.Argument(None, help="repository name"),
    dataset_type: str = typer.Option(
        "tdb2", help=f"dataset type. Options: {dataset_type_options}"
    ),
    config: Path | None = typer.Option(None, help="assembler file"),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    if dataset_name and config:
        raise typer.BadParameter("Only dataset name or --config is allowed, not both.")

    if dataset_name and dataset_type not in dataset_type_options:
        raise typer.BadParameter(
            f"Invalid dataset type '{dataset_type}'. Options: {dataset_type_options}"
        )

    if dataset_name:
        with httpx.Client(auth=auth, timeout=timeout) as client:
            try:
                result = create_dataset(fuseki_url, dataset_name, dataset_type, client)
                console.print(result)
            except Exception as err:
                console.print(
                    f"[bold red]ERROR[/bold red] Failed to create repository {dataset_name} at {fuseki_url}."
                )
                raise err
    else:
        if config is None:
            raise typer.BadParameter(
                "Either dataset name or assembler config file must be provided."
            )
        with httpx.Client(auth=auth, timeout=timeout) as client:
            try:
                result = create_dataset(fuseki_url, config, http_client=client)
                console.print(result)
            except Exception as err:
                console.print(
                    f"[bold red]ERROR[/bold red] Failed to create repository {dataset_name} at {fuseki_url}."
                )
                raise err


@app.command(name="upload", help="Upload files to a database repository")
def upload_command(
    path: Path = typer.Argument(
        ..., help="The path of a file or directory to be uploaded."
    ),
    fuseki_url: str = typer.Argument(
        ..., help="Repository SPARQL Endpoint URL. E.g. http://localhost:3030/ds"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
    disable_ssl_verification: Annotated[
        bool,
        typer.Option(
            "--disable-ssl-verification", "-k", help="Disable SSL verification."
        ),
    ] = False,
    host_header: Annotated[
        str | None, typer.Option("--host-header", "-h", help="Override the Host header")
    ] = None,
) -> None:
    """Upload a file or a directory of files with an RDF file extension.

    File extensions: [.nt, .nq, .ttl, .trig, .json, .jsonld, .xml]

    Files are uploaded into their own named graph in the format:
    <urn:file:{file.name}>
    E.g. <urn:file:example.ttl>
    """
    files = []

    if path.is_file():
        files.append(path)
    else:
        files += path.glob("**/*")

    auth = (
        (username, password) if username is not None and password is not None else None
    )

    files = list(filter(lambda f: f.suffix in suffix_map.keys(), files))

    headers = {}
    if host_header is not None:
        headers["Host"] = host_header

    with httpx.Client(
        auth=auth,
        timeout=timeout,
        headers=headers,
        verify=False if disable_ssl_verification else True,
    ) as client:
        for file in track(files, description=f"Uploading {len(files)} files..."):
            try:
                upload(fuseki_url, file, f"urn:file:{file.name}", http_client=client)
            except Exception as err:
                console.print(
                    f"[bold red]ERROR[/bold red] Failed to upload file {file}."
                )
                raise err


@app.command(name="clear", help="Clear a database repository")
def repository_clear_command(
    named_graph: str = typer.Argument(
        ..., help="Named graph. If 'all' is supplied, it will remove all named graphs."
    ),
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as client:
        try:
            clear_graph(fuseki_url, named_graph, client)
        except Exception as err:
            console.print(
                f"[bold red]ERROR[/bold red] Failed to run clear command with '{named_graph}' at {fuseki_url}."
            )
            raise err


@app.command(name="delete", help="Delete a database repository")
def repository_delete_command(
    fuseki_url: str = typer.Argument(
        ..., help="Fuseki base URL. E.g. http://localhost:3030"
    ),
    dataset_name: str = typer.Argument(..., help="The name of the dataset to delete."),
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
):
    auth = (
        (username, password) if username is not None and password is not None else None
    )

    with httpx.Client(auth=auth, timeout=timeout) as client:
        try:
            success_message = delete_dataset(fuseki_url, dataset_name, client)
            console.print(success_message)
        except FusekiError as err:
            console.print(err)


@app.command(name="sparql", help="Query a database repository")
def sparql_command3(
    path_or_url: Path = typer.Argument(
        ..., help="Repository SPARQL Endpoint URL. E.g. http://localhost:3030/ds"
    ),
    q: str = typer.Argument(..., help="The SPARQL query to sent to the database"),
    response_format: Annotated[
        str,
        typer.Option(
            "--response-format",
            "-f",
            help="The response format of the SPARQL query",
        ),
    ] = "table",
    username: Annotated[
        str, typer.Option("--username", "-u", help="Fuseki username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="Fuseki password.")
    ] = None,
    timeout: Annotated[
        int, typer.Option("--timeout", "-t", help="Timeout per request")
    ] = 60,
) -> None:
    sparql_command(path_or_url, q, response_format, username, password, timeout)



from pathlib import Path

from pyshacl import validate as v
from rdflib import Graph

from kurra.utils import load_graph


def validate(
    data_file_or_dir_or_graph: Path | Graph, shacl_file_or_dir_or_graph: Path | Graph
) -> tuple[bool, Graph, str]:
    """Runs pySHACL's validate() function with some preset values"""

    data_graph = load_graph(data_file_or_dir_or_graph)
    shapes_graph = load_graph(shacl_file_or_dir_or_graph)

    return v(data_graph, shacl_graph=shapes_graph, allow_warnings=True)



from typer.testing import CliRunner

from kurra.cli import app

runner = CliRunner()


def test_cli_db_list(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    result = runner.invoke(
        app,
        [
            "db",
            "list",
            f"http://localhost:{port}",
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert "'ds.name': '/ds'" in result.output



from pathlib import Path

import typer

from kurra.cli.console import console
from kurra.cli.utils import (
    format_shacl_graph_as_rich_table,
)
from kurra.shacl import validate

app = typer.Typer(help="SHACL commands")


@app.command(
    name="validate",
    help="Validate a given file or directory of RDF files using a given SHACL file or directory of files",
)
def shacl_command(
    file_or_dir: Path = typer.Argument(
        ..., help="The file or directory of RDF files to be validated"
    ),
    shacl_file_or_dir: Path = typer.Argument(
        ..., help="The file or directory of SAHCL files to validate with"
    ),
) -> None:
    """Validate a given file or directory of files using a given SHACL file or directory of files"""
    valid, g, txt = validate(file_or_dir, shacl_file_or_dir)
    if valid:
        console.print("The data is valid")
    else:
        console.print("The data is NOT valid")
        console.print("The errors are:")
        console.print(format_shacl_graph_as_rich_table(g))



import pytest

from kurra.db import FusekiError, list_dataset


def test_db_list_datasets(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}"
    return_value = list_dataset(base_url, http_client)
    assert "/ds" in list(map(lambda x: x["ds.name"], return_value))


def test_db_list_datasets_non_existent(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)
    base_url = f"http://localhost:{port}/some-url"
    with pytest.raises(FusekiError) as exc_info:
        list_dataset(base_url, http_client)

    assert (
        f"Failed to list datasets at http://localhost:{port}/some-url"
        in exc_info.value.message
    )



import json
from pathlib import Path

import pytest

from kurra.db import upload
from kurra.sparql import query
from kurra.utils import RenderFormat, render_sparql_result, make_httpx_client

LANG_TEST_VOC = Path(__file__).parent / "language-test.ttl"
TESTING_GRAPH = "https://example.com/testing-graph"


def test_query_db(fuseki_container, http_client):
    port = fuseki_container.get_exposed_port(3030)

    SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"
    upload(SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client)

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            GRAPH ?g {
                ?c a skos:Concept .
            } 
        }"""

    assert "--- | ---" in render_sparql_result(query(SPARQL_ENDPOINT, q, http_client))

    assert (
        "c"
        in json.loads(
            (
                render_sparql_result(
                    query(SPARQL_ENDPOINT, q, http_client), RenderFormat.json
                )
            )
        )["head"]["vars"]
    )

    # test return format options

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE {
            GRAPH ?g {
                ?c a skos:Concept ;
                    skos:prefLabel "English only"@en ;
                .
            }
        }"""

    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=True, return_bindings_only=True
    )
    assert r[0]["c"]["value"] == "https://example.com/demo-vocabs/language-test/en-only"

    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=True, return_bindings_only=False
    )
    assert (
        r["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=False, return_bindings_only=False
    )
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=False, return_bindings_only=True
    )
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2[0]["c"]["value"] == "https://example.com/demo-vocabs/language-test/en-only"
    )

    q = "ASK {?s ?p ?o}"
    r = query(SPARQL_ENDPOINT, q, http_client)  # False, False
    assert r == '{"head": {}, "boolean": true}'

    q = "ASK {?s ?p ?o}"
    r = query(SPARQL_ENDPOINT, q, http_client, return_python=True)
    assert r["boolean"]

    q = "ASK {?s ?p ?o}"
    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=True, return_bindings_only=True
    )
    assert r

    q = "ASK {?s ?p ?o}"
    r = query(SPARQL_ENDPOINT, q, http_client, return_bindings_only=True)
    assert r == "true"

    q = "ASK {?s ?p <http://nothing.com/x>}"
    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=True, return_bindings_only=True
    )
    assert not r


def test_query_file():
    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE {
            ?c a skos:Concept ;
                skos:prefLabel ?pl ;
            .
            
            OPTIONAL {
                ?c skos:altLabel ?al .
            }
        }
        LIMIT 3"""

    assert "--- | --- | ---" in render_sparql_result(query(LANG_TEST_VOC, q))

    assert (
        "pl"
        in json.loads(render_sparql_result(query(LANG_TEST_VOC, q), RenderFormat.json))[
            "head"
        ]["vars"]
    )

    # test return format options

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE {
            ?c a skos:Concept ;
                skos:prefLabel ?pl ;
            .

            OPTIONAL {
                ?c skos:altLabel ?al .
            }
        }
        LIMIT 3"""

    r = query(LANG_TEST_VOC, q, return_python=True, return_bindings_only=True)
    assert r[0]["c"]["value"] == "https://example.com/demo-vocabs/language-test/en-only"

    r = query(LANG_TEST_VOC, q, return_python=True, return_bindings_only=False)
    assert (
        r["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(LANG_TEST_VOC, q, return_python=False, return_bindings_only=False)
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(LANG_TEST_VOC, q, return_python=False, return_bindings_only=True)
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2[0]["c"]["value"] == "https://example.com/demo-vocabs/language-test/en-only"
    )

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q)  # False, False
    assert r == '{"head": {}, "boolean": true}'

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q, return_python=True)
    assert r["boolean"]

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q, return_python=True, return_bindings_only=True)
    assert r

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q, return_bindings_only=True)
    assert r == "true"

    q = "ASK {?s ?p <http://nothing.com/x>}"
    r = query(LANG_TEST_VOC, q, return_python=True, return_bindings_only=True)
    assert not r


def test_duplicates():
    rdf_data = """
    PREFIX people: <https://linked.data.gov.au/dataset/people/>
    PREFIX schema: <https://schema.org/>

    people:nick
        a
            schema:Person , 
            schema:Patient ;
        schema:name "Nick" ;
        schema:age 42 ;
        schema:parent people:george ;
    .

    people:george
        a schema:Person ; 
        schema:name "George" ;
        schema:age 70 ;    
    .
    """

    q = """
    PREFIX people: <https://linked.data.gov.au/dataset/people/>
    PREFIX schema: <https://schema.org/>

    SELECT ?p ?name
    WHERE {
        ?p 
            a schema:Person ;
            schema:name ?name ;
            schema:age ?age ;
        .

        FILTER (?age < 50)
    }
    """

    assert len(query(rdf_data, q, return_python=True, return_bindings_only=True)) == 1


def test_auth(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client)

    q = "ASK {?s ?p ?o}"
    r = query(
        SPARQL_ENDPOINT, q, http_client, return_python=True, return_bindings_only=True
    )
    assert r

    with pytest.raises(RuntimeError):
        r = query(
            SPARQL_ENDPOINT, q, None, return_python=True, return_bindings_only=True
        )


def test_construct(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client)

    q = """
        CONSTRUCT { ?s ?p ?o }
        WHERE {
            GRAPH ?g {
                ?s ?p ?o
            }
        }
        LIMIT 3       
        """

    r = query(SPARQL_ENDPOINT, q, http_client)
    assert len(r) == 3


def test_insert(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client)

    q = """
        INSERT { ?s ?p ?o }
        WHERE {
            GRAPH ?g {
                ?s ?p ?o
            }
        }
        LIMIT 3       
        """

    with pytest.raises(NotImplementedError):
        r = query(SPARQL_ENDPOINT, q, http_client)


from typer.testing import CliRunner

from kurra import __version__
from kurra.cli import app

runner = CliRunner()


def test_version_command():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert result.output == f"{__version__}\n"



from pathlib import Path

from typer.testing import CliRunner

from kurra.cli import app

runner = CliRunner()


def test_cli_db_create(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    result = runner.invoke(
        app,
        [
            "db",
            "create",
            f"http://localhost:{port}",
            "myds",
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert f"Dataset myds created at http://localhost:{port}." in result.output


def test_cli_db_create_with_both_dataset_name_and_config_file(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    current_dir = Path(__file__).parent
    file = current_dir / "config.ttl"
    result = runner.invoke(
        app,
        [
            "db",
            "create",
            f"http://localhost:{port}",
            "myds",
            "--config",
            str(file),
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 2
    assert "Only dataset name or --config is allowed, not both." in result.output


def test_cli_db_create_with_config_file(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    current_dir = Path(__file__).parent
    file = current_dir / "config.ttl"
    result = runner.invoke(
        app,
        [
            "db",
            "create",
            f"http://localhost:{port}",
            "--config",
            str(file),
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert (
        f"Dataset myds created using assembler config at http://localhost:{port}."
        in result.output
    )


def test_cli_db_create_existing_dataset(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    dataset_name = "ds"
    result = runner.invoke(
        app,
        [
            "db",
            "create",
            f"http://localhost:{port}",
            dataset_name,
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 1
    assert (
        f"Failed to create repository {dataset_name} at http://localhost:{port}"
        in result.output
    )



import json
from pathlib import Path

import httpx
from rdflib import BNode, Graph, Literal, URIRef

from kurra.db import sparql
from kurra.utils import load_graph, make_httpx_client


def query(
    p: Path | str | Graph,
    q: str,
    http_client: httpx.Client = None,
    return_python: bool = False,
    return_bindings_only: bool = False,
):
    if "CONSTRUCT" in q:
        if isinstance(p, str) and p.startswith("http"):
            if http_client is None:
                http_client = httpx.Client()

            headers = {
                "Content-Type": "application/sparql-query",
                "Accept": "text/turtle"
            }
            r = http_client.post(p, content=q, headers=headers)

            return Graph().parse(data=r.text, format="turtle")

        if isinstance(p, str) and not p.startswith("http"):
            # parse it and handle it as a Graph
            p = load_graph(p)

        if isinstance(p, Path):
            p = load_graph(p)

        # if we are here, path_str_graph_or_sparql_endpoint is a Graph
        r = p.query(q)
        g = Graph()
        for s, p, o in r:
            g.add((s, p, o))

        return g
    elif "INSERT" in q or "DELETE" in q:
        raise NotImplementedError("INSERT & DELETE queries are not yet implemented by this interface. Try kurra.db.sparql")

    else:  # SELECT or ASK
        close_http_client = False
        if http_client is None:
            http_client = httpx.Client()
            close_http_client = True

        r = None
        if isinstance(p, str) and p.startswith("http"):
            r = sparql(p, q, http_client, True, False)

        if r is None:
            r = {"head": {"vars": []}, "results": {"bindings": []}}
            x = load_graph(p).query(q)
            if x.vars is not None:
                for var in x.vars:
                    r["head"]["vars"].append(str(var))

                for row in x.bindings:
                    new_row = {}
                    for k in r["head"]["vars"]:
                        v = row.get(k)
                        if v is not None:
                            if isinstance(v, URIRef):
                                new_row[str(k)] = {"type": "uri", "value": str(v)}
                            elif isinstance(v, BNode):
                                new_row[str(k)] = {"type": "bnode", "value": str(v)}
                            elif isinstance(v, Literal):
                                val = {"type": "literal", "value": str(v)}
                                if v.language is not None:
                                    val["xml:lang"] = v.language
                                if v.datatype is not None:
                                    val["datatype"] = v.datatype
                                new_row[str(k)] = val

                    r["results"]["bindings"].append(new_row)
            else:
                r = {
                    "head": {},
                    "boolean": True if x.askAnswer else False,
                }

        if close_http_client:
            http_client.close()

        match (return_python, return_bindings_only):
            case (True, True):
                if r.get("results") is not None:
                    return r["results"]["bindings"]
                elif r.get("boolean") is not None:  # ASK
                    return r["boolean"]
                else:
                    return r
            case (True, False):
                return r
            case (False, True):
                if r.get("results") is not None:
                    return json.dumps(r["results"]["bindings"])
                elif r.get("boolean") is not None:  # ASK
                    return json.dumps(r["boolean"])
                else:
                    return json.dumps(r)
            case _:
                return json.dumps(r)



from io import TextIOBase
from pathlib import Path
from textwrap import dedent
from typing import Union

import httpx
from rdflib import RDF, Graph, URIRef

from kurra.utils import load_graph

suffix_map = {
    ".nt": "application/n-triples",
    ".nq": "application/n-quads",
    ".ttl": "text/turtle",
    ".trig": "application/trig",
    ".json": "application/ld+json",
    ".jsonld": "application/ld+json",
    ".xml": "application/rdf+xml",
}


class FusekiError(Exception):
    """An error that occurred while interacting with Fuseki."""

    def __init__(self, message_context: str, message: str, status_code: int) -> None:
        self.message = f"{status_code} {message_context}. {message}"
        super().__init__(self.message)


def _guess_query_is_update(query: str) -> bool:
    if any(x in query for x in ["DROP", "INSERT", "DELETE"]):
        return True
    else:
        return False


def _guess_return_type_for_sparql_query(query: str) -> str:
    if any(x in query for x in ["SELECT", "INSERT", "ASK"]):
        return "application/sparql-results+json"
    elif "CONSTRUCT" in query:
        return "text/turtle"
    else:
        return "application/sparql-results+json"


def upload(
    sparql_endpoint: str,
    file_or_str_or_graph: Union[Path, str, Graph],
    graph_id: str = None,
    append: bool = False,
    http_client: httpx.Client | None = None,
) -> None:
    """This function uploads a file to a SPARQL Endpoint using the Graph Store Protocol.

    It will upload it into a graph identified by graph_id (an IRI or Blank Node). If no graph_id is given, it will be
    uploaded into the Fuseki default graph.

    By default, it will replace all content in the Named Graph or default graph. If append is set to True, it will
    add it to existing content in the graph_id Named Graph.

    An httpx Client may be supplied for efficient client reuse, else each call to this function will recreate a new
    Client."""

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    params = {"graph": graph_id} if graph_id else "default"

    data = load_graph(file_or_str_or_graph).serialize(format="longturtle")
    headers = {"content-type": "text/turtle"}

    if append:
        response = http_client.post(sparql_endpoint, params=params, headers=headers, content=data)
    else:
        response = http_client.put(sparql_endpoint, params=params, headers=headers, content=data)

    status_code = response.status_code

    if status_code != 200 and status_code != 201 and status_code != 204:
        message = (
            str(file_or_str_or_graph)
            if isinstance(file_or_str_or_graph, Path)
            else "content"
        )
        raise RuntimeError(
            f"Received status code {status_code} for file {message} at url {sparql_endpoint}. Message: {response.text}"
        )

    if close_http_client:
        http_client.close()


def list_dataset(
    base_url: str,
    http_client: httpx.Client | None = None,
) -> dict:
    """
    List the datasets in a Fuseki server.

    :param base_url: The base URL of the Fuseki server. E.g., http://localhost:3030
    :param http_client: The synchronous httpx client to be used. If this is not provided, a temporary one will be created.
    :raises FusekiError: If the datasets fail to list or the server responds with an invalid data structure.
    :returns: The Fuseki listing of datasets as a dictionary.
    """
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    headers = {"accept": "application/json"}
    response = http_client.get(f"{base_url}/$/datasets", headers=headers)
    status_code = response.status_code

    if status_code != 200:
        raise FusekiError(
            f"Failed to list datasets at {base_url}", response.text, status_code
        )

    if close_http_client:
        http_client.close()

    try:
        datasets = response.json()["datasets"]
        return datasets
    except KeyError:
        raise FusekiError(
            f"Failed to parse datasets response from {base_url}",
            response.text,
            status_code,
        )


def create_dataset(
    sparql_endpoint: str,
    dataset_name_or_config_file: str | TextIOBase | Path,
    dataset_type: str = "tdb2",
    http_client: httpx.Client | None = None,
) -> str:
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    if isinstance(dataset_name_or_config_file, str):
        data = {"dbName": dataset_name_or_config_file, "dbType": dataset_type}
        response = http_client.post(f"{sparql_endpoint}/$/datasets", data=data)
        status_code = response.status_code
        if response.status_code != 200 and response.status_code != 201:
            raise FusekiError(
                f"Failed to create dataset {dataset_name_or_config_file} at {sparql_endpoint}",
                response.text,
                status_code,
            )
        msg = f"{dataset_name_or_config_file} created at"
    else:
        if isinstance(dataset_name_or_config_file, TextIOBase):
            data = dataset_name_or_config_file.read()
        else:
            with open(dataset_name_or_config_file, "r") as file:
                data = file.read()

        graph = Graph().parse(data=data, format="turtle")
        fuseki_service = graph.value(
            None, RDF.type, URIRef("http://jena.apache.org/fuseki#Service")
        )
        dataset_name = graph.value(
            fuseki_service, URIRef("http://jena.apache.org/fuseki#name")
        )

        response = http_client.post(
            f"{sparql_endpoint}/$/datasets",
            content=data,
            headers={"Content-Type": "text/turtle"},
        )
        status_code = response.status_code
        if response.status_code != 200 and response.status_code != 201:
            raise FusekiError(
                f"Failed to create dataset {dataset_name} at {sparql_endpoint}",
                response.text,
                status_code,
            )

        msg = f"{dataset_name} created using assembler config at"

    if close_http_client:
        http_client.close()

    return f"Dataset {msg} {sparql_endpoint}."


def clear_graph(
        sparql_endpoint: str,
        graph_id: str,
        http_client: httpx.Client):
    """
    Clears - remove all triples from - an identified graph or from all graphs if "all" is given as the graph_id
    """
    query = "CLEAR ALL" if graph_id == "all" else f"CLEAR GRAPH <{graph_id}>"
    headers = {"content-type": "application/sparql-update"}
    response = http_client.post(sparql_endpoint, headers=headers, content=query)
    status_code = response.status_code

    if status_code != 204:
        raise RuntimeError(
            f"Received status code {status_code}. Message: {response.text}"
        )


def delete_dataset(
    base_url: str, dataset_name: str, http_client: httpx.Client | None = None
) -> str:
    """
    Delete a Fuseki dataset.

    :param base_url: The base URL of the Fuseki server. E.g., http://localhost:3030
    :param dataset_name: The dataset to be deleted
    :param http_client: The synchronous httpx client to be used. If this is not provided, a temporary one will be created.
    :raises FusekiError: If the dataset fails to delete.
    :returns: A message indicating the successful deletion of the dataset.
    """
    if not dataset_name:
        raise ValueError("You must supply a dataset name")

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    response = http_client.delete(f"{base_url}/$/datasets/{dataset_name}")
    status_code = response.status_code

    if status_code != 200:
        raise FusekiError(
            f"Failed to delete dataset '{dataset_name}'", response.text, status_code
        )

    if close_http_client:
        http_client.close()

    return f"Dataset {dataset_name} deleted."


def sparql(
    sparql_endpoint: str,
    query: str,
    http_client: httpx.Client = None,
    return_python: bool = False,
    return_bindings_only: bool = False,
):
    """Poses a SPARQL query to a SPARQL Endpoint"""

    if http_client is None:
        http_client = httpx.Client()

    if query is None:
        raise ValueError("You must supply a query")

    if _guess_query_is_update(query):
        headers = {"Content-Type": "application/sparql-update"}
    else:
        headers = {"Content-Type": "application/sparql-query"}

    headers["Accept"] = _guess_return_type_for_sparql_query(query)

    response = http_client.post(
        sparql_endpoint,
        headers=headers,
        content=query,
    )

    status_code = response.status_code

    if status_code != 200 and status_code != 201 and status_code != 204:
        raise RuntimeError(f"ERROR {status_code}: {response.text}")

    match (return_python, return_bindings_only):
        case (True, True):
            return response.json()["results"]["bindings"]
        case (True, False):
            return response.json()
        case (False, True):
            return dedent(response.text.split('"bindings": [')[1].split("]")[0])
        case _:
            return response.text



from pathlib import Path

import httpx

from kurra.db import clear_graph, sparql, upload


def test_file_upload_ng_replacement(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    with httpx.Client() as client:
        SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
        minimalist_rdf = Path(__file__).parent / "minimal3.ttl"

        upload(
            SPARQL_ENDPOINT,
            minimalist_rdf,
            "https://example.com/demo-vocabs/language-test",
            False,
            client,
        )


def test_query(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    with httpx.Client() as client:
        SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
        TESTING_GRAPH = "https://example.com/testing-graph"

        data = """
                PREFIX ex: <http://example.com/>
                
                ex:a ex:b ex:c .
                ex:a2 ex:b2 ex:c2 .
                """

        upload(SPARQL_ENDPOINT, data, TESTING_GRAPH, False, client)

        q = """
            SELECT (COUNT(*) AS ?count) 
            WHERE {
              GRAPH <XXX> {
                ?s ?p ?o
              }
            }        
            """.replace("XXX", TESTING_GRAPH)

        r = sparql(
            SPARQL_ENDPOINT, q, client, return_python=True, return_bindings_only=True
        )

        count = int(r[0]["count"]["value"])

        assert count == 2

        q = "DROP GRAPH <XXX>".replace("XXX", TESTING_GRAPH)

        print("QUERY")
        print(q)
        print("QUERY")

        r = sparql(SPARQL_ENDPOINT, q, client)

        print(r)


def test_clear(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    with httpx.Client() as client:
        SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
        TESTING_GRAPH = "https://example.com/testing-graph"

        data = """
                PREFIX ex: <http://example.com/>

                ex:a ex:b ex:c .
                ex:a2 ex:b2 ex:c2 .
                """

        upload(SPARQL_ENDPOINT, data, TESTING_GRAPH, False, client)

        q = """
            SELECT (COUNT(*) AS ?count) 
            WHERE {
              GRAPH <XXX> {
                ?s ?p ?o
              }
            }        
            """.replace("XXX", TESTING_GRAPH)

        r = sparql(
            SPARQL_ENDPOINT, q, client, return_python=True, return_bindings_only=True
        )

        count = int(r[0]["count"]["value"])

        assert count == 2

        clear_graph(SPARQL_ENDPOINT, TESTING_GRAPH, client)

        r = sparql(
            SPARQL_ENDPOINT, q, client, return_python=True, return_bindings_only=True
        )

        count = int(r[0]["count"]["value"])

        assert count == 0
